require 'json'
require 'uu/os/lang/json_patch'
require 'uu/os/server/abstract_servlet'
require 'uu/os/server/parameter_parser'
require 'uu/os/cmd/command_client'
require 'uu/os/cmd/command_executor'
require 'uu/os/ipc/ipc_error'
require 'uu/os/ipc/error_kind'
require 'uu/os/security/session'
require 'uu/os/logger'
require 'uu/os/cmd/async_command_state'
require 'uu/os/util/url_converter'

module UU
  module OS
    module Util

      # Test implementation of API for invocation of asynchronous commands
      # and API for obtaining asynchronous command result. This implementation
      # is used only in case application server is started in development
      # mode (so that it is possible to test asynchronous commands).
      # @private
      class EmbeddedCommandExecutor < UU::OS::Server::AbstractServlet # :nodoc:
        include UU::OS::Server::ParameterParser
        include UU::OS::Util::URLConverter

        PATH = '/ues/wcp/ues/platform/command/UESAsyncCommandExecutor'

        HEALTH_PATH = '(/)?_uu_/health'

        ASYNC_CMD_TIMEOUT = 3600

        CALLBACK_TIMEOUT = 5

        @@task_storage = {}

        @@callback_storage = {}

        @@task_register = {}

        @@base_url = nil

        # Registers embedded executor for server in development mode.
        # @private
        def self.on_server_startup(ctx) # :nodoc:          
          return if !ctx.development_mode? && ENV['LOCAL_EXECUTOR'] !~ /^true$/
          ctx.register_servlet(self, true, PATH)
          @@base_url = Thread.current[:BASE_URL] || ENV['LOCAL_EXECUTOR_BASE_URL']
          path_prefix = UU::OS::Util::URLConverter.concat_path(ENV['LOCAL_EXECUTOR_PATH_PREFIX'], 'ues/wcp')
          if !@@base_url.nil?
            # Configure command client to invoke executor locally.
            url_cfg_key = 'uu.os.command.uesasynccommandexecutor.server-url'
            prefix_cfg_key = 'uu.os.command.uesasynccommandexecutor.server-path-prefix'
            ctx.configuration[url_cfg_key] = @@base_url
            ctx.configuration[prefix_cfg_key] = path_prefix
            UU::OS::CMD::CommandClient.load_configuration(url_cfg_key => @@base_url, prefix_cfg_key => path_prefix)
          end
          # Patch command client to be able to hook callbacks
          # from thirdparty asynchronous commands.
          UU::OS::CMD::CommandClient.class_variable_set(:@@test_server_url, @@base_url)
          if !UU::OS::CMD::CommandClient.private_instance_methods.include?(:__invoke_orig)
            UU::OS::CMD::CommandClient.send(:alias_method, :__invoke_orig, :__invoke)
            UU::OS::CMD::CommandClient.send(:define_method, :__invoke) do |cmd_url_path, action, uu_uri, orig_parameters|
              # we need to duplicate params because extract_callback method delete uuCallback from parameters so callbacks
              # do not work in tests
              if !orig_parameters.nil?
                parameters = orig_parameters.dup
                parameters[:parameters] = parameters[:parameters].dup if !parameters[:parameters].nil?
              end
              callback = extract_callback(parameters)
              # In case we are invoking command on other than local server and callback is requested...
              # TODO Why direct access via "@@test_server_url" does not work?
              # It works if class is fully reopened but does not work in method defined via "define_method".
              if self.class.class_variable_get(:@@test_server_url) != @http_client.instance_variable_get(:@base_url) && !callback.nil?
                # Do not modify original parameters
                parameters = parameters.dup
                # Delete callback so that it is not used by remote server
                # (where invocation will fail)
                parameters.delete(:callback)
                result = __invoke_orig(cmd_url_path, action, uu_uri, parameters)
                if result.kind_of?(UU::OS::Lang::Future)
                  # Register returned future to invoke callback once it is finished
                  UU::OS::Util::EmbeddedCommandExecutor.handle_external_callback(result, callback)
                end
                return result
              else
                return __invoke_orig(cmd_url_path, action, uu_uri, orig_parameters)
              end
            end
          end
        end

        # Processes incoming request.
        # @private
        def service(request) # :nodoc:
          Thread.current[:start] = Time.now
          begin
            UU::OS::Security::Session.login(request.headers[:authorization][0]) if request.headers.contains(:authorization)
            method = request.url.split('/')[-1]
            if method =~ /invoke/
              return response(405, nil, UU::OS::IPC::IPCError.new(UU::OS::IPC::ErrorKind::CLIENT, nil, "Invalid invocation method, expected POST.").to_json) if request.method != :post
              invoke(request)
            elsif method =~ /getState/
              return response(405, nil, UU::OS::IPC::IPCError.new(UU::OS::IPC::ErrorKind::CLIENT, nil, "Invalid invocation method, expected GET.").to_json) if request.method != :get
              get_state(request)
            else
              return response(405, nil, UU::OS::IPC::IPCError.new(UU::OS::IPC::ErrorKind::CLIENT, nil, "Resource #{request.url} does not exist.").to_json)
            end
          ensure
            UU::OS::Security::Session.logout(true)
          end
        end

        # Invokes asynchronous part of command according to incoming
        # parameters (execution is done in a separate thread, therefore
        # it is really asynchronous). As result, a random generated task
        # URI is returned.
        # @private
        def invoke(request) # :nodoc:
          parameters = parse_parameters(request)
          queue = parameters.delete(:uesuri)
          return response(400, nil, UU::OS::IPC::IPCError.new(UU::OS::IPC::ErrorKind::CLIENT, 'UU.OS/E05R04.M01', "Main entity is not set.").to_json) if queue.nil? || queue.empty?
          main_object_uri = parameters.delete(:mainObjectUri)
          return response(400, nil, UU::OS::IPC::IPCError.new(UU::OS::IPC::ErrorKind::CLIENT, 'UU.OS/E05R04.M01', "Parameter mainObjectUri must be set").to_json) if main_object_uri.nil? || main_object_uri.empty?
          app_url_code = parameters[:appUrlCode]
          return response(400, nil, UU::OS::IPC::IPCError.new(UU::OS::IPC::ErrorKind::CLIENT, 'UU.OS/E05R04.M01', "Parameter appUrlCode must be set").to_json) if app_url_code.nil? || app_url_code.empty?
          cmd_url_path = parameters[:cmdUrlPath]
          return response(400, nil, UU::OS::IPC::IPCError.new(UU::OS::IPC::ErrorKind::CLIENT, 'UU.OS/E05R04.M01', "Parameter cmdUrlPath must be set").to_json) if cmd_url_path.nil? || cmd_url_path.empty?
          action = "#{parameters[:action]}Async"
          if parameters[:callback].kind_of?(Hash)
            cbauc = parameters[:callback][:appUrlCode]
            return response(400, nil, UU::OS::IPC::IPCError.new(UU::OS::IPC::ErrorKind::CLIENT, 'UU.OS/E05R04.M01', "Parameter callback.appUrlCode must be set").to_json) if cbauc.nil? || cbauc.empty?
            cbcup = parameters[:callback][:cmdUrlPath]
            return response(400, nil, UU::OS::IPC::IPCError.new(UU::OS::IPC::ErrorKind::CLIENT, 'UU.OS/E05R04.M01', "Parameter callback.cmdUrlPath must be set").to_json) if cbcup.nil? || cbcup.empty?
            cba = parameters[:callback][:action]
          end

          session = UU::OS::Security::Session.current_session
          task_uri = "ues:[#{rand(100000..999999)}]:[#{rand(100000..999999)}]:[#{rand(100000..999999)}]"
          start = Thread.current[:start]

          if "#{cmd_url_path}/#{action}" =~ /#{HEALTH_PATH}/
            sync_inv = {cont: 'EXECUTOR'}
            sync_inv[:nodeId] = get_node_id(request)
            sync_inv[:timestamp] = start
          end

          @@task_register[task_uri] = concat_path(app_url_code, cmd_url_path, action)
          @@task_storage[task_uri] = Thread.new do
            start = Time.now
            if "#{cmd_url_path}/#{action}" =~ /#{HEALTH_PATH}/
              health = [sync_inv]
              invocation = {cont: 'ASYNC_EXECUTOR'}
              invocation[:nodeId] = get_node_id(request)
              invocation[:timestamp] = start
              health << invocation
            end
            begin
              session = session.login(parameters[:authToken]) if !parameters[:authToken].nil?
              client = UU::OS::CMD::CommandClient.new(app_url_code, session, request_timeout: ASYNC_CMD_TIMEOUT, base_url: @@base_url)
              result = client.send(action, cmd_url_path, main_object_uri, parameters: parameters[:parameters])
              if "#{cmd_url_path}/#{action}" =~ /#{HEALTH_PATH}/
                health << result
                result = health
              end
            rescue Exception => e
              log_error('Asynchronous command failed.', e)
              e.instance_variable_set(:@remote, false)
              result = e
              if "#{cmd_url_path}/#{action}" =~ /#{HEALTH_PATH}/
                invocation[:exception] = e.to_json
                result = health
              end
            end
            if parameters[:callback]
              parameters[:callback][:mainObjectUri] ||= main_object_uri
              @@callback_storage[task_uri] = callback(request, session, parameters[:callback], result)
            end
            if "#{cmd_url_path}/#{action}" =~ /#{HEALTH_PATH}/
              invocation[:duration] = ((Time.now - start) * 1_000).to_i
              store_result(task_uri, result, true)
            else
              store_result(task_uri, result)
            end
          end
          response(200, {content_type: 'application/json'}, task_uri.to_json)
        end

        # Returns current state of the invoked asynchronous command.
        # @private
        def get_state(request) # :nodoc:
          parameters = parse_parameters(request)
          main_object_uri = parameters.delete(:uesuri)

          if @@task_register[main_object_uri].nil?
            return response(404, nil, UU::OS::IPC::IPCError.new(UU::OS::IPC::ErrorKind::CLIENT, 'UU.OS/E05R03.M05', "Execution task #{main_object_uri} does not exist.").to_json)
          end
          result = @@task_storage[main_object_uri]
          if result.kind_of?(Thread)
            state = {state: UU::OS::CMD::AsyncCommandState::RUNNING, result: nil}
          elsif result.kind_of?Exception
            state = {state: UU::OS::CMD::AsyncCommandState::ERROR, exception: result.to_json}
          else
            state = {state: UU::OS::CMD::AsyncCommandState::FINISHED, result: result}
          end
          if @@callback_storage.has_key?(main_object_uri)
            cbt_uri = @@callback_storage[main_object_uri]
            state[:callbackTaskUri] = cbt_uri
            cb_res = @@task_storage[cbt_uri]
            if cb_res.kind_of?(Thread)
              state[:callbackState] = UU::OS::CMD::AsyncCommandState::RUNNING
            elsif cb_res.kind_of?Exception
              state[:callbackState] = UU::OS::CMD::AsyncCommandState::ERROR
              state[:callbackException] = cb_res.to_json
            else
              state[:callbackState] = UU::OS::CMD::AsyncCommandState::FINISHED
              state[:callbackResult] = cb_res
            end
          end
          response(200, {content_type: 'application/json'}, state.to_json)
        end

        # Invokes callback method.
        # @private
        def callback(request, session, callback, result) # :nodoc:
          start = Time.now
          app_url_code = callback[:appUrlCode]
          cmd_url_path = callback[:cmdUrlPath]
          action = callback[:action]
          task_uri = "ues:[#{rand(100000..999999)}]:[#{rand(100000..999999)}]:[#{rand(100000..999999)}]"

          if "#{cmd_url_path}/#{action}" =~ /#{HEALTH_PATH}/
            sync_inv = {cont: 'EXECUTOR'}
            sync_inv[:nodeId] = get_node_id(request)
            sync_inv[:timestamp] = start
          end

          @@task_register[task_uri] = concat_path(app_url_code, cmd_url_path, action)
          @@task_storage[task_uri] = Thread.new do
            start = Time.now
            if "#{cmd_url_path}/#{action}" =~ /#{HEALTH_PATH}/
              health = [sync_inv]
              invocation = {cont: 'ASYNC_EXECUTOR'}
              invocation[:nodeId] = get_node_id(request)
              invocation[:timestamp] = start
              health << invocation
            end
            begin
              session = session.login(callback[:authToken]) if !callback[:authToken].nil?
              parameters = callback[:parameters] || {}
              parameters[:uuCtx] = {}
              parameters[:uuCtx][:result] = {}
              if result.kind_of?(Exception)
                parameters[:uuCtx][:result][:state] = UU::OS::CMD::AsyncCommandState::ERROR
                parameters[:uuCtx][:result][:exception] = result.to_json
              else
                parameters[:uuCtx][:result][:state] = UU::OS::CMD::AsyncCommandState::FINISHED
                parameters[:uuCtx][:result][:result] = result
              end
              client = UU::OS::CMD::CommandClient.new(app_url_code, session, request_timeout: CALLBACK_TIMEOUT, base_url: @@base_url)
              result = client.send(action, cmd_url_path, callback[:mainObjectUri], parameters: parameters)
              if "#{cmd_url_path}/#{action}" =~ /#{HEALTH_PATH}/
                health << result
                invocation[:duration] = ((Time.now - start) * 1_000).to_i
                store_result(task_uri, health, true)
              else
                store_result(task_uri, result)
              end
            rescue Exception => e
              log_error('Callback failed.', e)
              e.instance_variable_set(:@remote, false)
              if "#{cmd_url_path}/#{action}" =~ /#{HEALTH_PATH}/
                invocation[:exception] = e.to_json
                invocation[:duration] = ((Time.now - start) * 1_000).to_i
                store_result(task_uri, health, true)
              else
                store_result(task_uri, e)
              end
            end
          end
          return task_uri
        end

        # Stores command result.
        # @private
        def store_result(task_uri, result, allow_struct = false) # :nodoc:
          if result.nil? || result.kind_of?(String) || result.kind_of?(Exception) || allow_struct
            @@task_storage[task_uri] = result
          elsif result.kind_of?(Hash) && result.has_key?(:uri)
            @@task_storage[task_uri] = result[:uri]
          else
            @@task_storage[task_uri] = nil
            log_error("Object of type #{result.class} cannot be stored as execution result, only URI is supported.")
          end
        end

        # Logs using {UU::OS::Logger}.
        # @private
        def log_error(message, exception = nil) # :nodoc:
          message, exception = exception, message if message.kind_of?Exception
          log = UU::OS::Logger.new(self.class)
          log.error(message, exception)
        end

        # Returns node ID.
        # @private
        def get_node_id(request) # :nodoc:
          node_id = ENV['UU_NODE_ID']
          if node_id.nil? && Object.const_defined?(:Java)
            node_id = Java::JavaLang::System.getProperty('uu.node.id')
          end
          if node_id.nil?
            node_id = request.headers[:host][0]
          end
          return node_id
        end

        # Methods emulates callback from external service by checking result
        # of asynchronous command and invoking callback once command is finished.
        # @param [UU::OS::Lang::Future] future Instance of future for asynchronous command.
        # @param [Object] callback Callback parameters.
        # @private
        def self.handle_external_callback(future, callback) # :nodoc:
          # Extract task URI and executor from future to be able to check state of
          # asynchronous command directly without any processing of result (which
          # is not needed because we need raw data to be passed to callback). We
          # also need to bypass maximum timeout which is 30 seconds for future.
          task_uri = future.instance_variable_get(:@actor).instance_variable_get(:@task_uri)
          executor = future.instance_variable_get(:@actor).instance_variable_get(:@client)
          cmd_url_path = future.instance_variable_get(:@actor).instance_variable_get(:@cmd_url_path)
          main_entity_uri = future.instance_variable_get(:@actor).instance_variable_get(:@main_entity_uri)
          cb_task_uri = "ues:[#{rand(100000..999999)}]:[#{rand(100000..999999)}]:[#{rand(100000..999999)}]"
          @@task_register[cb_task_uri] = UU::OS::Util::URLConverter.concat_path(callback[:appUrlCode], callback[:cmdUrlPath], callback[:action])
          @@task_storage[cb_task_uri] = Thread.new(task_uri, cb_task_uri, executor, cmd_url_path, main_entity_uri, callback) do |task_uri, cb_task_uri, executor, cmd_url_path, main_entity_uri, callback|
            begin
              state = executor.exec_async_status(cmd_url_path, main_entity_uri, parameters: {taskUri: task_uri})
              wait = 0.5
              while state[:state] == UU::OS::CMD::AsyncCommandState::WAITING || state[:state] == UU::OS::CMD::AsyncCommandState::RUNNING do
                # Use same patter for waiting on external asynchronous services as in future
                # in order not to flood thirdparty server with get state commands.
                wait *= 2
                sleep(wait)
                state = executor.exec_async_status(cmd_url_path, main_entity_uri, parameters: {taskUri: task_uri})
              end
              parameters = callback[:parameters] || {}
              parameters[:uuCtx] = {}
              parameters[:uuCtx][:result] = {}
              parameters[:uuCtx][:result][:state] = state[:state]
              if state[:state] == UU::OS::CMD::AsyncCommandState::FINISHED
                parameters[:uuCtx][:result][:result] = state[:result]
              else
                parameters[:uuCtx][:result][:exception] = state[:exception]
              end
              if callback[:authToken].nil?
                session = executor.instance_variable_get(:@session)
              else
                UU::OS::Security::Session.login(callback[:authToken])
                session = UU::OS::Security::Session.current_session
              end
              client = UU::OS::CMD::CommandClient.new(callback[:appUrlCode], session, request_timeout: CALLBACK_TIMEOUT, base_url: @@base_url)
              begin
                result = client.send(callback[:action], callback[:cmdUrlPath], callback[:mainObjectUri], parameters: parameters)
                @@task_storage[cb_task_uri] = result
              rescue Exception => cbe
                @@task_storage[cb_task_uri] = cbe
              end
            rescue Exception => e
              # Since this error can be caused only by testing environment,
              # print problem to standard output to be immediately visible.
              msg = "ERROR #{Time.now} Callback cannot be invoked.\n"
              msg << "#{e.class} [#{e.id}]: #{e.message}"
              trace = e.backtrace
              msg << "\n  " << trace.join("\n  ") if !trace.nil? && !trace.empty?
              STDERR.puts(msg)
              @@task_storage[cb_task_uri] = e
            end
          end
          @@callback_storage[task_uri] = cb_task_uri
        end

      end

    end
  end
end

# Extend test class with method for waiting for asynchronous callback.
# @private
module MiniTest # :nodoc:
  class Test

    # Waits for finishing of all asynchronous tasks (asynchronous commands and callbacks)
    # running on local test server.
    # @param [Fixnum] timeout Defines how long to wait (defaults to 1 minute).
    # @return [Hash] Hash containing results of all local tasks (command URL is used as
    #    key, value contains particular result or instance of exception).
    def wait_for_async_tasks(timeout = 60)
      Timeout::timeout(timeout) do
        finished = false
        while !finished do
          sleep(0.5)
          result = {}
          finished = true
          UU::OS::Util::EmbeddedCommandExecutor.class_variable_get(:@@task_storage).each do |key, value|
            if value.kind_of?(Thread)
              finished = false
              break
            end
            result[UU::OS::Util::EmbeddedCommandExecutor.class_variable_get(:@@task_register)[key]] = value
          end
        end
        return result
      end
    end

    # Clears local storage of asynchronous tasks. This method should be invoked between each
    # test method (or each asynchronous command invocation) to be sure result provided by
    # {MiniTest::Test#wait_for_async_tasks} is actual.
    def clear_async_tasks
      UU::OS::Util::EmbeddedCommandExecutor.class_variable_get(:@@task_storage).clear
      UU::OS::Util::EmbeddedCommandExecutor.class_variable_get(:@@task_register).clear
      UU::OS::Util::EmbeddedCommandExecutor.class_variable_get(:@@callback_storage).clear
    end

  end
end
